
#%%
def add(a: list, b: list) -> list:
    """Add two numbers represented as lists of digits in the decimal system."""
    # check the number of digits and create an equally long list of 
    # output digits that initially contains zeros
    l = len(a)
    sum = [0]*l

    # we set initially set the "carry digit" to zero
    carry = 0

    # since the algorithm must move from right to left, i.e. starting at the 
    # least significant digit, we must start at the index l-1 and decrease the index 
    # in each step until we have reached the left-most digit with index zero.
    i = l-1
    while i >= 0:

        # sum the current digits and account for carry digit
        s = a[i] + b[i] + carry

        # the digit of the sum must be from 0 to 9
        # if the sum is ten or larger, we add a 1 for the carry 
        # and reduce the sum by 10     
        if s>=10:
            carry = 1
            s = s - 10
        else: # otherwise we reset the carry for the next step
            carry = 0
        # set the digit for the current position in the sum
        sum[i] = s

        # continue with the next position
        i = i-1
    return sum

#%% first example
# as input, we pass two lists of decimal digits, where the most significant (i.e. largest) digit is the first, i.e. left-most entry in the list. 
# to simplify the implementation, we require to always have a leading zero, such that the output list has the same length as the input lists
a = [0,0,2,1]
b = [0,0,2,1]

# we now execute our algorithm on these inputs and print the returned output
print(a)
print(b)
s = add(a, b)
print(s)

#%% example with carry digit
a = [0,0,2,1]
b = [0,0,2,9]

print('a =', a)
print('b =', b)
s = add(a, b)
print('s =', s)

#%% example where the output has one bit more than the inputs
a = [0,9,9,9]
b = [0,0,0,1]

print('a =', a)
print('b =', b)
s = add(a, b)
print('s =', s)

#%% note that this yields a wrong result! Why?
a = [9,9,9]
b = [0,0,1]

print('a =', a)
print('b =', b)
s = add(a, b)
print('s =', s)

#%% example for two numbers that cannot be added by the ALU of a 64-bit processor
a = [0,5,6,3,2,5,7,8,2,1,9,5,6,3,2,5,7,8,2,1,9,9,9]
b = [0,5,6,3,2,5,7,8,2,1,9,5,6,3,2,5,7,8,2,1,9,9,9]

s = add(a, b)
print('s =', s)

#%% does this work using the + operator in python?
a = 5632578219563257821999
b = 5632578219563257821999
s = a + b
print('s =', s)

# %%
