#%%
import string 
from collections import Counter
import pandas as pd

alphabet = list(string.ascii_lowercase)
alphabet.append(' ')

def vigenere_table(key, reverse=False):
    table = {}
    for k in key:
        if reverse:
            table[k] = { alphabet[i]: alphabet[(i+k)%len(alphabet)] for i in range(len(alphabet))}
        else:
            table[k] = { alphabet[(i+k)%len(alphabet)]: alphabet[i] for i in range(len(alphabet))}
    return table

def vigenere_encrypt(plain: str, key: list) -> str:
    cipher = []
    table = vigenere_table(key, reverse=False)

    i = 0
    for c in plain:
        # get shift for current position
        k = key[i%len(key)]
        cipher.append(table[k][c])
        i += 1    
    return ''.join(cipher)

def vigenere_decrypt(cipher: str, key: list) -> str:
    plain = []
    table = vigenere_table(key, reverse=True)
    i = 0
    for c in cipher:
        # get shift for current position
        k = key[i%len(key)]
        plain.append(table[k][c])
        i += 1
    return ''.join(plain)


#%% Alice and Bob share secret key 
secret_key = [24, 7, 12, 7, 0, 2, 17]

#%% This key allows to build a Vigenere table that we can use to encrypt and decrypt
vigenere_table(secret_key)

#%% Alice encrypts message and send it to Bob
text = """To be or not to be that is the question
Whether tis nobler in the mind to suffer
The slings and arrows of outrageous fortune
Or to take arms against a sea of troubles
And by opposing end them To die to sleep
No more and by a sleep to say we end
The heart ache and the thousand natural shocks
That flesh is heir to tis a consummation
Devoutly to be wishd To die to sleep
To sleep perchance to dreamay theres the rub
For in that sleep of death what dreams may come
When we have shuffled off this mortal coil
Must give us pause theres the respect
That makes calamity of so long life
For who would bear the whips and scorns of time
The oppressors wrong the proud mans contumely
The pangs of disprizd love the laws delay
The insolence of office and the spurns
That patient merit of the unworthy takes
When he himself might his quietus make
With a bare bodkin Who would fardels bear
To grunt and sweat under a weary life
But that the dread of something after death
The undiscovered country  from whose bourn
No traveller returns puzzles the will
And makes us rather bear those ills we have
Than fly to others that we know not of
Thus conscience doth make cowards of us all
And thus the native hue of resolution
Is sicklied over with the pale cast of thought
And enterprises of great pith and moment
With this regard their currents turn awry
And lose the name of action""".lower().replace('\n', ' ')

cipher = vigenere_encrypt(text, secret_key)
print('Alice sends: \t', cipher)

# Bob receives cipher and uses key to decrypt it
plain = vigenere_decrypt(cipher, secret_key)
print('Bob decrypts: \t', plain)


#%% Carol intercepts message and analyses character frequencies
char_freq_cipher = Counter(cipher)
df = pd.DataFrame.from_dict(char_freq_cipher, orient='index', columns=['cipher'])
df = df.sort_values(axis=0, by=['cipher'], ascending=False)
df.plot.bar()

#%% we further assume that Carol has access to character frequencies in common English texts
char_freq_plain = Counter(text)
df = pd.DataFrame.from_dict(char_freq_plain, orient='index', columns=['plain'])
df = df.sort_values(axis=0, by=['plain'], ascending=False)
df.plot.bar()



# %%
